@@ -4,93 +4,15 @@ Classes for Numerosity analysis.
 from tifffile import TiffFile, imread, imsave
 import numpy as np
 import json
+import os
 
 import matplotlib.pyplot as plt
 import warnings
 from tqdm import tqdm
 import pandas as pd
+import PyPDF2
 
-
-def extract_windows(array, window_size):
-    """
-    breaks ND array into windows along 0 dimention, starting from the begining and till the end.
-    """
-    start = 0
-    num_windows = array.shape[0] - window_size + 1
-
-    sub_windows = (
-            start +
-            np.expand_dims(np.arange(window_size), 0) +
-            np.expand_dims(np.arange(num_windows), 0).T
-    )
-    return array[sub_windows]
-
-
-def get_dff(array, window_size):
-    """
-    subtracts average baseline from an ND array along the 0 dimention.
-    window_size must be an odd number.
-
-    The baseline for the first and the last window//2 elements is the same :
-    the first or the last value calculated with the full window
-    """
-    percentile = 8  # 8th percentile
-    mean_signal = np.percentile(extract_windows(array, window_size), percentile, axis=1)
-    # construct the baseline
-    start = window_size // 2
-    end = len(array) - window_size // 2
-    # add the beginning and end to baseline
-    baseline = np.zeros(array.shape)
-    baseline[0:start] = mean_signal[0]
-    baseline[start:end] = mean_signal
-    baseline[end:] = mean_signal[-1]
-
-    return (array - baseline) / baseline, start, end
-
-
-def get_t_score(movie1, movie2, absolute=False):
-    """
-    Returns absolute t-score image ( 2D or 3D, depending on input),
-    for t-score calculations see for example :
-    https://www.jmp.com/en_us/statistics-knowledge-portal/t-test/two-sample-t-test.html
-    """
-    # TODO : check input dimensions
-
-    avg1 = np.mean(movie1, axis=0)
-    avg2 = np.mean(movie2, axis=0)
-
-    var1 = np.var(movie1, axis=0)
-    var2 = np.var(movie2, axis=0)
-    n1 = movie1.shape[0]
-    n2 = movie2.shape[0]
-
-    std = np.sqrt(((n1 - 1) * var1 + (n2 - 1) * var2) / (n1 + n2 - 2))
-
-    if absolute:
-        t_score = np.absolute(avg1 - avg2) / std
-    else:
-        t_score = (avg1 - avg2) / std
-
-    return t_score
-
-
-def get_diff(movie1, movie2, absolute=False):
-    """
-    Returns absolute difference image ( 2D or 3D, depending on input).
-    per pixel: calculates the mean image for each movie, subtracts and takes the absolute value.
-    """
-    # TODO : check input dimensions
-
-    avg1 = np.mean(movie1, axis=0)
-    avg2 = np.mean(movie2, axis=0)
-
-    if absolute:
-        diff = np.absolute(avg1 - avg2)
-    else:
-        diff = (avg1 - avg2)
-
-    return diff
-
+from .utils import *
 
 class Spot:
     """
@@ -162,9 +84,25 @@ class Spot:
     def __repr__(self):
         return self.__str__()
 
-    def mask_at_zero(self):
+    def mask_at_zero(self, diameter=None, units='pix'):
+        """
+        diameter : list or numpy array or int, diameter in zyx order. If int, then it is a sphere.
+                    If not None, diameter will be used to create a mask.
+                    If None, then spots diameter will be used and units will be ignored.
+        units: what units is the diameter in : 'pix' or 'phs'
+        """
+
+        if diameter is None:
+            d = self.diameter['pix']
+        else:
+            if isinstance(diameter, int):
+                diameter = np.array([1, 1, 1]) * diameter
+            else:
+                diameter = np.array(diameter)
+            if units == 'phs':
+                diameter = np.round(diameter / self.resolution)
+            d = diameter
 
-        d = self.diameter['pix']
         r = d / 2  # radius
         # make sure diameter is odd: round to the next odd number
         d = ((d // 2) * 2 + 1).astype(np.int)
@@ -188,10 +126,14 @@ class Spot:
 
         return mask, centered_idx
 
-    def create_mask(self, volumes):
+    def create_mask(self, volumes, diameter=None, units='pix'):
         """
         Creates a binary mask that can be applied to the volume.
-        volumes : sequence of volumes in shape TZYX or one volume in ZYX
+        volumes : image as a sequence of volumes in shape TZYX or one volume in ZYX
+        diameter : list or numpy array or int, diameter in zyx order. If int, then it is a sphere.
+                    If not None, diameter will be used to create a mask.
+                    If None, then spots diameter will be used and units will be ignored.
+        units: what units is the diameter in : 'pix' or 'phs'
         """
         # create mask for a single volume
         if len(volumes.shape) == 4:
@@ -199,11 +141,11 @@ class Spot:
         elif len(volumes.shape) == 3:
             zmax, ymax, xmax = volumes.shape
         else:
-            raise AssertionError("volumes should be 4D tzyx or 3d zyx")
+            raise AssertionError("volumes should be 4D tzyx or 3D zyx")
         mask = np.zeros((zmax, ymax, xmax))
 
         # get list of nonzero elements , centered at zero
-        _, idx = self.mask_at_zero()
+        _, idx = self.mask_at_zero(diameter=diameter, units=units)
 
         # shift to center
         shift = self.center['pix']
@@ -266,7 +208,8 @@ class Signals:
         return Signals(dff, traces_type="dff")
 
     @classmethod
-    def from_spots(cls, spots, volumes=None, experiment=None, batch_size=None, movie=None, traces_type="raw"):
+    def from_spots(cls, spots, volumes=None, experiment=None, batch_size=None, movie=None,
+                   verbose=False, traces_type="raw"):
         """
         Extracts signals from volumes for each spot in spots.
         spots : list[Spot]
@@ -290,6 +233,7 @@ class Signals:
 
         N = len(spots)
 
+        design_matrix = None
         # if volumes are given explicitly as a movie
         if movie is not None:
             T = movie.shape[0]
@@ -300,7 +244,6 @@ class Signals:
         elif volumes is not None and experiment is not None:
             T = len(volumes)
             design_matrix = np.zeros((T, N))
-
             # split volumes into batches of certain size
             if batch_size is not None:
                 btcs = np.array_split(volumes, np.ceil(T / batch_size))
@@ -310,9 +253,11 @@ class Signals:
             t_start = 0
             for batch in btcs:
                 t_end = t_start + len(batch)
-                print(f"Batch {batch}, t_start {t_start}, t_end {t_end}")
-
-                movie = experiment.volume_manager.load_volumes(batch)
+                if verbose:
+                    print(f"Batch {batch}, t_start {t_start}, t_end {t_end}")
+                # load a batch from disk
+                movie = experiment.volume_manager.load_volumes(batch, show_progress=True)
+                # extract signals for all the spots from the loaded chunk
                 design_matrix = fill_signal(design_matrix, movie, t_start, t_end)
 
                 t_start = t_end
@@ -321,17 +266,9 @@ class Signals:
 
     def get_looped(self, trace, experiment, time_points=None, error_type="prc"):
         """
-
-        Parameters
-        ----------
-        error_type
-        trace : trace id
-        experiment
-        time_points
-
-        Returns
-        -------
-
+        Returns signals looped per cycle
+        time_points: time points of the cycle. If you only need certain time-points from the cycle ( in volumes )
+        if you need to stack timepoints for each cycle, for example: [[0,1,2],[7,8,9]], pass the time_points as 2D list.
         """
 
         n_cycles = experiment.full_cycles
@@ -344,17 +281,32 @@ class Signals:
                                                     "by which I mean that plot_looped function needs more code :), " \
                                                     "Sorry. Hint : use time_points "
         cycled = self.traces[:, trace].reshape((n_cycles, int(cycle_length)))
+
+        # crop out the necessary cycle part,
+        # note: you can reorder the cycle as well ( to add points from the beginning  to the end of the cycle)
+        # you can also overlay some cycle time intervals, useful for psh
+        if time_points is not None:
+            time_points = np.array(time_points)
+            time_shape = time_points.shape
+            assert len(time_shape) < 3, "time_shape should be 1D or 2D"
+            cycled = cycled[:, time_points.flatten()]
+            # if you want to overlay cycle intervals:
+            if len(time_shape) == 2:
+                cycled = cycled.reshape(-1, time_shape[1])
+
         mean = np.mean(cycled, axis=0)
 
         if error_type == "prc":
             # error bars : 5 to 95 th percentile around the median
             e = np.r_[np.expand_dims(mean - np.percentile(cycled, 5, axis=0), axis=0),
                       np.expand_dims(np.percentile(cycled, 95, axis=0) - mean, axis=0)]
-        if error_type == "sem":
+        elif error_type == "sem":
             # error bars : sem around hte mean
             sem = np.std(cycled, axis=0, ddof=1) / np.sqrt(cycled.shape[0])
             e = np.r_[np.expand_dims(sem, axis=0),
                       np.expand_dims(sem, axis=0)]
+        else:
+            e = None
 
         return cycled, mean, e
 
@@ -378,16 +330,38 @@ class Spots:
         """
 
         self.spots = spots
+        self.num_spots = len(self.spots)
         self.signals = signals
         # TODO : have signals as dict with " raw" , " dff" etc
         self.groups = None
         if groups is not None:
             self.add_groups(groups)
 
-    def add_groups(self, groups):
+    def __str__(self):
+
+        groups = "No"
+        if self.groups is not None:
+            groups = self.groups.keys()
+
+        signals = "not loaded"
+        if self.signals is not None:
+            signals = "loaded"
+
+        return f"{self.num_spots} spots\n " \
+               f"{groups} groups\n" \
+               f"Signals {signals}"
+
+    def __repr__(self):
+        return self.__str__()
+
+    def add_groups(self, groups, rewrite = False):
         """
         groups: dict with boolean arrays to say which cells belong.
         """
+        if not rewrite and self.groups is not None:
+            for group in groups:
+                assert group not in self.groups, f"The group {group} already exists," \
+                                                 f" use rewrite to overwrite it, or use another group name"
         if self.groups is None:
             self.groups = {}
         for group in groups:
@@ -396,7 +370,8 @@ class Spots:
 
     def list_groups(self):
         if self.groups is not None:
-            print(self.groups.keys())
+            for key in self.groups.keys():
+                print(key)
 
     def _get_centers(self, units='phs'):
         """
@@ -434,7 +409,6 @@ class Spots:
         if self.groups is not None:
             j_dict["groups"] = {}
             for group in self.groups:
-                print(group)
                 j_dict["groups"][group] = np.array(self.groups[group], dtype=bool).tolist()
 
         if self.signals is not None:
@@ -490,15 +464,25 @@ class Spots:
 
         return cls(spots, groups=None, signals=None)
 
-    def get_group_mask(self, group, mask_shape):
+    def get_signals(self, volumes=None, experiment=None, batch_size=None, movie=None, traces_type="raw", reload=False):
+        assert reload or self.signals is None, "Spots already have signals loaded, reload? "
+
+        self.signals = Signals.from_spots(self.spots, volumes=volumes, experiment=experiment,
+                                          batch_size=batch_size, movie=movie, traces_type=traces_type)
+
+    def get_group_mask(self, group, mask_shape, diameter=None, units='pix'):
         """
         Create a 3D volume that only shows the spots that belong to the particular group
         group : list[bool], the length of spots
         mask_shape : the shape of the 3D volume of the mask
+        diameter : list or numpy array or int, diameter in zyx order. If int, then it is a sphere.
+                    If not None, diameter will be used to create a mask.
+                    If None, then spots diameter will be used and units will be ignored.
+        units: what units is the diameter in : 'pix' or 'phs'
         """
         mask = np.zeros(mask_shape)
         for ispot in tqdm(np.where(group)[0]):
-            imask, _ = self.spots[ispot].create_mask(mask)
+            imask, _ = self.spots[ispot].create_mask(mask, diameter=diameter, units=units)
             mask = mask + imask
         return mask
 
@@ -509,128 +493,42 @@ class Spots:
         traces = self.signals.traces[:, group]
         return Signals(traces, traces_type=self.signals.traces_type)
 
-    def get_signals(self, volumes=None, experiment=None, batch_size=None, movie=None, traces_type="raw", reload=False):
-        assert reload or self.signals is None, "Spots already have signals loaded, reload? "
-
-        self.signals = Signals.from_spots(self.spots, volumes=volumes, experiment=experiment,
-                                          batch_size=batch_size, movie=movie, traces_type=traces_type)
-
-
-class SignalPlotter:
-
-    def __init__(self, signals, experiment, spf=1):
+    def get_group_idx(self, group):
         """
-
-        Parameters
-        ----------
-        signals
-        experiment
-        spf : seconds per frame
+        Returns indices for a particular group only
         """
-        self.signals = signals
-        self.experiment = experiment
-        self.n_signals = self.signals.traces.shape[1]
+        idx = np.arange(self.num_spots)
+        idx = idx[group]
+        return idx
 
-    def plot_labels(self, ax, extent=None, front_to_tail=None):
-        # timing in volumes
-        timing = (self.experiment.cycle.timing / self.experiment.volume_manager.fpv).astype(np.int)
-        conditions = [cond for t, condition in zip(timing, self.experiment.cycle.conditions) for cond in
-                      [condition.name] * t]
-        # return_inverse gives the integer encoding
-        names, values = np.unique(conditions, return_inverse=True)
-        if front_to_tail is not None:
-            old_order = np.arange(len(values))
-            new_order = np.r_[old_order[front_to_tail:], old_order[0:front_to_tail]]
-            values = values[new_order]
-
-        img = ax.imshow(values[np.newaxis, :], aspect='auto',
-                        extent=extent, cmap=plt.get_cmap('Greys', len(names)))
-
-        return names, values, img
-
-    def show_labels(self, x_step=1):
+    def get_group_centers(self, group, units='pix'):
         """
-        Keep in mind - assign colors in alphabetic order of the condition name.
+        Returns centers for a particular group only
+        units: the units of the center to return
         """
+        centers = []
+        for ispot, spot in enumerate(self.spots):
+            if group[ispot]:
+                centers.append(spot.center[units])
 
-        fig = plt.figure(figsize=(10, 5))
-        ax = fig.add_subplot(111)
-        names, values, img = self.plot_labels(ax)
-        plt.xticks(ticks=np.arange(0, len(values), x_step))
-        plt.xlabel('volume # per cycle')
-        plt.title('Stimulus cycle')
-        ax.get_yaxis().set_visible(False)
+        return np.array(centers)
 
-        # TODO : for now it fits 3 different colors only! fix it!
-        cbar = plt.colorbar(img, ax=ax, ticks=[0.5, 1, 1.5], orientation='horizontal')
-        cbar.ax.set_xticklabels(names)
-
-    def show_psh(self, traces, main_title, tittles, error_type="prc", plot_individual=True,
-                 front_to_tail=None, figure_layout=None, figsize=None,
-                 ylabel='', xlabel='', noise_color='--c', dpi=160):
+    def get_group_info(self, group_list, group = None):
         """
-        front_to_tail : how many cycle points to attach from front to tail
+        Returns a string with the titles of the groups from group_list, where each spot is a member.
+        group = the T/F list ... If group is not None: only returns it for the group.
         """
-
-        def plot_errorbar(ax, mean, e):
-            ax.errorbar(np.arange(len(mean)), mean, yerr=e, fmt='o', color='r')
-            ax.plot(mean, color='r')
-
-        def get_ax_limits(cycled, mean, e, plot_individual):
-            if plot_individual:
-                ymin = np.min(cycled)
-                ymax = np.max(cycled)
-            else:
-                ymin = np.min(mean - e[1, :])
-                ymax = np.max(mean + e[0, :])
-            xmin = -0.5
-            xmax = cycled.shape[1] - 0.5
-
-            return xmin, xmax, ymin, ymax
-
-        if figure_layout is not None:
-            n_rows = figure_layout[0]
-            n_col = figure_layout[1]
-        else:
-            n_rows = len(traces)
-            n_col = 1
-
-        if figsize is None:
-            figsize = (12, n_rows * 4)
-
-        fig, axes = plt.subplots(n_rows, n_col, figsize=figsize, dpi=dpi)
-        axes = axes.flatten()
-        fig.suptitle(main_title)
-        for plot_id, trace in enumerate(traces):
-            cycled, mean, e = self.signals.get_looped(trace, self.experiment, error_type=error_type)
-
-            if front_to_tail is not None:
-                old_order = np.arange(len(mean))
-                new_order = np.r_[old_order[front_to_tail:], old_order[0:front_to_tail]]
-
-                cycled = cycled[:, new_order]
-                mean = mean[new_order]
-                e = e[:, new_order]
-
-            ax = axes[plot_id]
-            xmin, xmax, ymin, ymax = get_ax_limits(cycled, mean, e, plot_individual)
-            names, _, img = self.plot_labels(ax, extent=[xmin, xmax, ymin, ymax], front_to_tail=front_to_tail)
-
-            if plot_individual:
-                ax.plot(cycled.T, noise_color, alpha=0.3)
-
-            plot_errorbar(ax, mean, e)
-
-            ax.set_title(tittles[plot_id])
-            ax.set_xlim((xmin, xmax))
-            ax.set_ylim((ymin, ymax))
-            ax.set_xticks(np.arange(len(mean)))
-            ax.set_xticklabels([])
-            ax.set_ylabel(ylabel)
-            ax.set_xlabel(xlabel)
-
-        # cbar = plt.colorbar(img, ax=ax, ticks=[0.5, 1.5, 2.5], orientation='horizontal')
-        # cbar.ax.set_xticklabels(names)
+        group_info = []
+        for ispot, spot in enumerate(self.spots):
+            groups = ""
+            for group_name in group_list:
+                if self.groups[group_name][ispot]:
+                    groups = groups + group_name + "; "
+            group_info.append(groups)
+        group_info = np.array(group_info)
+        if group is not None:
+            group_info = group_info[group]
+        return group_info
 
 
 class SignalAnalyzer:
@@ -683,7 +581,7 @@ class SignalAnalyzer:
         data2 = self.signals.traces[group2, :]
         T1, nspots = data1.shape
         T2 = data1.shape[0]
-        print(f" Data points in group 1 : {T1},in group 1 : {T2}.\nNumber of spots : {nspots}.")
+        # print(f" Data points in group 1 : {T1},in group 1 : {T2}.\nNumber of spots : {nspots}.")
 
         p = []
         for ispot in tqdm(np.arange(nspots)):
@@ -772,3 +670,5 @@ class Preprocess:
             # exit cycle the first time you saw the end og the experiment
             if chunk[-1] == (n_volumes - 1):
                 break
+
+
